from dataclasses import dataclass
from enum import Enum
from typing import Any, cast

from graphql import (
    EnumValueNode,
    GraphQLArgument,
    GraphQLEnumType,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLType,
    InputValueDefinitionNode,
    ListValueNode,
    ObjectValueNode,
)


class GraphQLEnumsValuesVisitor:
    schema: GraphQLSchema
    enum_values: dict[str, dict[str, Any]]

    def __init__(self, schema: GraphQLSchema):
        self.enum_values = {}
        self.schema = schema

        self.visit_enum_types()
        self.visit_schema()

    def visit_enum_types(self) -> None:
        for type_def in self.schema.type_map.values():
            if isinstance(type_def, GraphQLEnumType):
                self.enum_values[type_def.name] = {
                    key: value.value for key, value in type_def.values.items()
                }

    def visit_schema(self) -> None:
        raise NotImplementedError(
            "GraphQLEnumsValuesVisitor subclasses must implement 'visit_schema'"
        )


class GraphQLSchemaEnumsValuesVisitor(GraphQLEnumsValuesVisitor):
    def visit_schema(self) -> None:
        for type_def in self.schema.type_map.values():
            if type_def.name.startswith("__"):
                continue  # Skip introspection types

            if isinstance(type_def, GraphQLObjectType | GraphQLInterfaceType):
                self.visit_object(type_def)

            if isinstance(type_def, GraphQLInputObjectType):
                self.visit_input(type_def)

    def visit_object(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType,
    ) -> None:
        for field_name, field_def in object_def.fields.items():
            for arg_name, arg_def in field_def.args.items():
                self.visit_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                )

    def visit_input(
        self,
        input_def: GraphQLInputObjectType,
    ) -> None:
        for field_name, field_def in input_def.fields.items():
            self.visit_value(input_def, field_name, field_def)

    def visit_value(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType | GraphQLInputObjectType,
        field_name: str,
        field_def: GraphQLField | GraphQLInputField,
        arg_name: str | None = None,
        arg_def: GraphQLArgument | None = None,
    ) -> None:
        src_def: GraphQLInputField | GraphQLArgument
        if isinstance(field_def, GraphQLInputField):
            src_def = field_def
        elif isinstance(arg_def, GraphQLArgument):
            src_def = arg_def

        if src_def.default_value:
            if is_graphql_list(src_def.type) and isinstance(
                src_def.default_value, list
            ):
                self.visit_list_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    src_def.type,
                    src_def.default_value,
                )

            else:
                src_type = unwrap_type(src_def.type)
                if isinstance(src_type, GraphQLEnumType) and is_raw_enum_value(
                    src_def.default_value
                ):
                    self.visit_schema_enum_default_value(
                        GraphQLSchemaEnumDefaultValueLocation(
                            enum_name=src_type.name,
                            enum_value=src_def.default_value,
                            object_name=object_def.name,
                            object_def=object_def,
                            field_name=field_name,
                            field_def=field_def,
                            arg_name=arg_name,
                            arg_def=arg_def,
                            default_value=src_def.default_value,
                            default_value_path=None,
                        )
                    )

                elif isinstance(src_type, GraphQLInputObjectType):
                    self.visit_input_value(
                        object_def,
                        field_name,
                        field_def,
                        arg_name,
                        arg_def,
                        src_type,
                        src_def.default_value,
                    )

    def visit_list_value(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType | GraphQLInputObjectType,
        field_name: str,
        field_def: GraphQLField | GraphQLInputField,
        arg_name: str | None,
        arg_def: GraphQLArgument | None,
        value_def: GraphQLType,
        value: Any,
    ) -> None:
        value_type = unwrap_list_type(value_def)
        if is_graphql_list(value_type):
            for value_item in value:
                self.visit_list_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    value_type,
                    value_item,
                )

        elif isinstance(value_type, GraphQLEnumType):
            for default_value_path, enum_value in enumerate(value):
                if is_raw_enum_value(enum_value):
                    self.visit_schema_enum_default_value(
                        GraphQLSchemaEnumDefaultValueLocation(
                            enum_name=value_type.name,
                            enum_value=enum_value,
                            object_name=object_def.name,
                            object_def=object_def,
                            field_name=field_name,
                            field_def=field_def,
                            arg_name=arg_name,
                            arg_def=arg_def,
                            default_value=value,
                            default_value_path=default_value_path,
                        )
                    )

        elif isinstance(value_type, GraphQLInputObjectType):
            for value_item in value:
                if isinstance(value_item, dict):
                    self.visit_input_value(
                        object_def,
                        field_name,
                        field_def,
                        arg_name,
                        arg_def,
                        value_type,
                        value_item,
                    )

    def visit_input_value(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType | GraphQLInputObjectType,
        field_name: str,
        field_def: GraphQLField | GraphQLInputField,
        arg_name: str | None,
        arg_def: GraphQLArgument | None,
        value_def: GraphQLInputObjectType,
        value: dict,
    ) -> None:
        for input_field_name, input_field_def in value_def.fields.items():
            input_field_value = value.get(input_field_name)
            if input_field_value is None:
                continue  # Skip field

            input_field_type = unwrap_type(input_field_def.type)
            if is_graphql_list(input_field_def.type) and isinstance(
                input_field_value, list
            ):
                self.visit_list_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    input_field_def.type,
                    input_field_value,
                )

            elif isinstance(input_field_type, GraphQLEnumType) and is_raw_enum_value(
                input_field_value
            ):
                self.visit_schema_enum_default_value(
                    GraphQLSchemaEnumDefaultValueLocation(
                        enum_name=input_field_type.name,
                        enum_value=input_field_value,
                        object_name=object_def.name,
                        object_def=object_def,
                        field_name=field_name,
                        field_def=field_def,
                        arg_name=arg_name,
                        arg_def=arg_def,
                        default_value=value,
                        default_value_path=input_field_name,
                    )
                )

            elif isinstance(input_field_type, GraphQLInputObjectType) and isinstance(
                input_field_value, dict
            ):
                self.visit_input_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    input_field_type,
                    input_field_value,
                )

    def visit_schema_enum_default_value(
        self, location: "GraphQLSchemaEnumDefaultValueLocation"
    ) -> None:
        pass


@dataclass(frozen=True)
class GraphQLSchemaEnumDefaultValueLocation:
    enum_name: str
    enum_value: Any
    object_name: str
    object_def: GraphQLObjectType | GraphQLInterfaceType | GraphQLInputObjectType
    field_name: str
    field_def: GraphQLField | GraphQLInputField
    arg_name: str | None = None
    arg_def: GraphQLArgument | None = None
    default_value: Any = None
    default_value_path: Any = None


class GraphQLASTEnumsValuesVisitor(GraphQLEnumsValuesVisitor):
    def visit_schema(self) -> None:
        for type_def in self.schema.type_map.values():
            if type_def.name.startswith("__"):
                continue  # Skip introspection types

            if (
                isinstance(type_def, GraphQLObjectType | GraphQLInterfaceType)
                and type_def.ast_node
            ):
                self.visit_object(type_def)

            if isinstance(type_def, GraphQLInputObjectType) and type_def.ast_node:
                self.visit_input(type_def)

    def visit_object(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType,
    ) -> None:
        for field_name, field_def in object_def.fields.items():
            for arg_name, arg_def in field_def.args.items():
                if arg_def.ast_node and arg_def.ast_node.default_value:
                    self.visit_value(
                        object_def,
                        field_name,
                        field_def,
                        arg_name,
                        arg_def,
                    )

    def visit_input(
        self,
        input_def: GraphQLInputObjectType,
    ) -> None:
        for field_name, field_def in input_def.fields.items():
            if field_def.ast_node and field_def.ast_node.default_value:
                self.visit_value(
                    input_def,
                    field_name,
                    field_def,
                )

    def visit_value(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType | GraphQLInputObjectType,
        field_name: str,
        field_def: GraphQLField | GraphQLInputField,
        arg_name: str | None = None,
        arg_def: GraphQLArgument | None = None,
    ) -> None:
        src_def: GraphQLInputField | GraphQLArgument
        if isinstance(field_def, GraphQLInputField):
            src_def = field_def
        elif isinstance(arg_def, GraphQLArgument):
            src_def = arg_def

        ast_node = cast(InputValueDefinitionNode, src_def.ast_node)
        default_value_ast = ast_node.default_value
        if is_graphql_list(src_def.type) and isinstance(
            default_value_ast, ListValueNode
        ):
            self.visit_list_value(
                object_def,
                field_name,
                field_def,
                arg_name,
                arg_def,
                src_def.type,
                default_value_ast,
            )

        else:
            src_type = unwrap_type(src_def.type)
            if isinstance(src_type, GraphQLEnumType) and isinstance(
                default_value_ast, EnumValueNode
            ):
                self.visit_ast_enum_default_value(
                    GraphQLASTEnumDefaultValueLocation(
                        enum_name=src_type.name,
                        enum_value=default_value_ast.value,
                        object_name=object_def.name,
                        object_def=object_def,
                        field_name=field_name,
                        field_def=field_def,
                        arg_name=arg_name,
                        arg_def=arg_def,
                        ast_node=default_value_ast,
                        ast_node_path=None,
                    )
                )

            elif isinstance(src_type, GraphQLInputObjectType) and isinstance(
                default_value_ast, ObjectValueNode
            ):
                self.visit_input_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    src_type,
                    default_value_ast,
                )

    def visit_list_value(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType | GraphQLInputObjectType,
        field_name: str,
        field_def: GraphQLField | GraphQLInputField,
        arg_name: str | None,
        arg_def: GraphQLArgument | None,
        value_def: GraphQLType,
        value_ast: ListValueNode,
    ) -> None:
        value_type = unwrap_list_type(value_def)
        if is_graphql_list(value_type):
            for value_item in value_ast.values:
                self.visit_list_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    value_type,
                    cast(ListValueNode, value_item),
                )

        elif isinstance(value_type, GraphQLEnumType):
            for default_value_path, enum_value in enumerate(value_ast.values):
                if isinstance(enum_value, EnumValueNode):
                    self.visit_ast_enum_default_value(
                        GraphQLASTEnumDefaultValueLocation(
                            enum_name=value_type.name,
                            enum_value=enum_value.value,
                            object_name=object_def.name,
                            object_def=object_def,
                            field_name=field_name,
                            field_def=field_def,
                            arg_name=arg_name,
                            arg_def=arg_def,
                            ast_node=value_ast,
                            ast_node_path=default_value_path,
                        )
                    )

        elif isinstance(value_type, GraphQLInputObjectType):
            for value_item in value_ast.values:
                if isinstance(value_item, ObjectValueNode):
                    self.visit_input_value(
                        object_def,
                        field_name,
                        field_def,
                        arg_name,
                        arg_def,
                        value_type,
                        value_item,
                    )

    def visit_input_value(
        self,
        object_def: GraphQLObjectType | GraphQLInterfaceType | GraphQLInputObjectType,
        field_name: str,
        field_def: GraphQLField | GraphQLInputField,
        arg_name: str | None,
        arg_def: GraphQLArgument | None,
        value_def: GraphQLInputObjectType,
        value_ast: ObjectValueNode,
    ) -> None:
        value: dict[str, Any] = {
            field.name.value: field.value for field in value_ast.fields
        }

        for input_field_name, input_field_def in value_def.fields.items():
            input_field_value = value.get(input_field_name)
            if input_field_value is None:
                continue  # Skip field

            input_field_type = unwrap_type(input_field_def.type)
            if is_graphql_list(input_field_def.type) and isinstance(
                input_field_value, ListValueNode
            ):
                self.visit_list_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    input_field_def.type,
                    input_field_value,
                )

            elif isinstance(input_field_type, GraphQLEnumType) and isinstance(
                input_field_value, EnumValueNode
            ):
                self.visit_ast_enum_default_value(
                    GraphQLASTEnumDefaultValueLocation(
                        enum_name=input_field_type.name,
                        enum_value=input_field_value.value,
                        object_name=object_def.name,
                        object_def=object_def,
                        field_name=field_name,
                        field_def=field_def,
                        arg_name=arg_name,
                        arg_def=arg_def,
                        ast_node=value,
                        ast_node_path=input_field_name,
                    )
                )

            elif isinstance(input_field_type, GraphQLInputObjectType) and isinstance(
                input_field_value, ObjectValueNode
            ):
                self.visit_input_value(
                    object_def,
                    field_name,
                    field_def,
                    arg_name,
                    arg_def,
                    input_field_type,
                    input_field_value,
                )

    def visit_ast_enum_default_value(
        self, location: "GraphQLASTEnumDefaultValueLocation"
    ) -> None:
        pass


@dataclass(frozen=True)
class GraphQLASTEnumDefaultValueLocation:
    enum_name: str
    enum_value: Any
    object_name: str
    object_def: GraphQLInputObjectType | GraphQLInterfaceType | GraphQLObjectType
    field_name: str
    field_def: GraphQLField | GraphQLInputField
    arg_name: str | None = None
    arg_def: GraphQLArgument | None = None
    ast_node: Any = None
    ast_node_path: Any = None


def unwrap_type(graphql_type: GraphQLType) -> GraphQLType:
    if isinstance(graphql_type, GraphQLList | GraphQLNonNull):
        return unwrap_type(cast(GraphQLType, graphql_type.of_type))

    return graphql_type


def unwrap_list_type(graphql_type: GraphQLType) -> GraphQLType:
    if isinstance(graphql_type, GraphQLNonNull):
        return unwrap_list_type(cast(GraphQLType, graphql_type.of_type))

    if isinstance(graphql_type, GraphQLList):
        return unwrap_nonnull_type(cast(GraphQLType, graphql_type.of_type))

    return graphql_type


def unwrap_nonnull_type(graphql_type: GraphQLType) -> GraphQLType:
    if isinstance(graphql_type, GraphQLNonNull):
        return unwrap_nonnull_type(cast(GraphQLType, graphql_type.of_type))

    return graphql_type


def is_graphql_list(graphql_type: GraphQLType) -> bool:
    if isinstance(graphql_type, GraphQLNonNull):
        return is_graphql_list(cast(GraphQLType, graphql_type.of_type))

    return isinstance(graphql_type, GraphQLList)


def is_raw_enum_value(value: Any) -> bool:
    # Raw enum value is a str with key name
    return isinstance(value, str) and not isinstance(value, Enum)



from ariadne.contrib.relay.arguments import (
    ConnectionArguments,
)
from ariadne.contrib.relay.connection import RelayConnection
from ariadne.contrib.relay.objects import (
    RelayNodeInterfaceType,
    RelayObjectType,
    RelayQueryType,
)
from ariadne.contrib.relay.types import ConnectionResolver, GlobalIDTuple
from ariadne.contrib.relay.utils import decode_global_id, encode_global_id

__all__ = [
    "ConnectionArguments",
    "RelayNodeInterfaceType",
    "RelayConnection",
    "RelayObjectType",
    "RelayQueryType",
    "ConnectionResolver",
    "GlobalIDTuple",
    "decode_global_id",
    "encode_global_id",
]



from reprlib import repr
from traceback import format_exception
from typing import cast

from graphql import GraphQLError

from .utils import unwrap_graphql_error


def format_error(error: GraphQLError, debug: bool = False) -> dict:
    """Format the GraphQL error into JSON serializable format.

    If `debug` is set to `True`, error's JSON will also include the `extensions`
    key with `exception` object containing error's `context` and `stacktrace`.

    Returns a JSON-serializable `dict` with error representation.

    # Required arguments

    `error`: an `GraphQLError` to convert into JSON serializable format.

    # Optional arguments

    `debug`: a `bool` that controls if debug data should be included in
    result `dict`. Defaults to `False`.
    """

    formatted = cast(dict, error.formatted)
    if debug:
        if "extensions" not in formatted:
            formatted["extensions"] = {}
        formatted["extensions"]["exception"] = get_error_extension(error)
    return formatted


def get_error_extension(error: GraphQLError) -> dict | None:
    """Get a JSON-serializable `dict` containing error's stacktrace and context.

    Returns a JSON-serializable `dict` with `stacktrace` and `context` to include
    under error's `extensions` key in JSON response. Returns `None` if `error`
    has no stacktrace or wraps no exception.

    # Required arguments

    `error`: an `GraphQLError` to return context and stacktrace for.
    """
    unwrapped_error = unwrap_graphql_error(error)
    if unwrapped_error is None or not error.__traceback__:
        return None

    return {
        "stacktrace": get_formatted_error_traceback(unwrapped_error),
        "context": get_formatted_error_context(unwrapped_error),
    }


def get_formatted_error_traceback(error: Exception) -> list[str]:
    """Get JSON-serializable stacktrace from `Exception`.

    Returns a list of strings, with every item being a separate line
    from the stacktrace.

    This approach produces better results in GraphQL explorers, which display every
    line under previous one but not always format linebreak characters for blocks
    of text.

    # Required arguments

    `error`: an `Exception` to return formatted stacktrace for.
    """
    formatted = []
    for line in format_exception(type(error), error, error.__traceback__):
        formatted.extend(line.rstrip().splitlines())
    return formatted


def get_formatted_error_context(error: Exception) -> dict | None:
    """Get JSON-serializable context from `Exception`.

    Returns a `dict` of strings, with every key being value name and value
    being `repr()` of it's Python value. Returns `None` if context is not
    available.

    # Required arguments

    `error`: an `Exception` to return formatted context for.
    """

    tb_last = error.__traceback__
    while tb_last and tb_last.tb_next:
        tb_last = tb_last.tb_next
    if tb_last is None:
        return None
    return {key: repr(value) for key, value in tb_last.tb_frame.f_locals.items()}



from collections.abc import Sequence
from typing import Any

from ariadne.contrib.relay.arguments import ConnectionArgumentsUnion


class RelayConnection:
    def __init__(
        self,
        edges: Sequence[Any],
        total: int,
        has_next_page: bool,
        has_previous_page: bool,
        id_field: str = "id",
    ) -> None:
        self.edges = edges
        self.total = total
        self.has_next_page = has_next_page
        self.has_previous_page = has_previous_page
        self.id_field = id_field

    def get_cursor(self, obj):
        return obj[self.id_field]

    def get_node(self, obj):
        return obj

    def get_page_info(self, connection_arguments: ConnectionArgumentsUnion):
        return {
            "hasNextPage": self.has_next_page,
            "hasPreviousPage": self.has_previous_page,
            "startCursor": self.get_cursor(self.edges[0]) if self.edges else None,
            "endCursor": self.get_cursor(self.edges[-1]) if self.edges else None,
        }

    def get_edges(self):
        return [
            {"node": self.get_node(obj), "cursor": self.get_cursor(obj)}
            for obj in self.edges
        ]



from collections.abc import Awaitable
from typing import Any


class Explorer:
    def html(self, request: Any) -> str | None | Awaitable[str | None]:
        raise NotImplementedError("Explorer subclasses should define 'html' method")


class ExplorerHttp405(Explorer):
    """Explorer that always returns HTTP 405 not allowed response."""

    def html(self, request: Any) -> None:
        return None



import enum
from typing import (
    Any,
    cast,
)

from graphql.type import GraphQLEnumType, GraphQLNamedType, GraphQLSchema

from .types import SchemaBindable


class EnumType(SchemaBindable):
    """Bindable mapping Python values to enumeration members in a GraphQL schema.

    # Example

    Given following GraphQL enum:

    ```graphql
    enum UserRole {
        MEMBER
        MODERATOR
        ADMIN
    }
    ```

    You can use `EnumType` to map it's members to Python `Enum`:

    ```python
    user_role_type = EnumType(
        "UserRole",
        {
            "MEMBER": 0,
            "MODERATOR": 1,
            "ADMIN": 2,
        }
    )
    ```

    `EnumType` also works with dictionaries:

    ```python
    user_role_type = EnumType(
        "UserRole",
        {
            "MEMBER": 0,
            "MODERATOR": 1,
            "ADMIN": 2,
        },
    )
    ```
    """

    def __init__(
        self,
        name: str,
        values: dict[str, Any] | type[enum.Enum] | type[enum.IntEnum],
    ) -> None:
        """Initializes the `EnumType` with `name` and `values` mapping.

        # Required arguments

        `name`: a `str` with the name of GraphQL enum type in GraphQL schema to
        bind to.

        `values`: a `dict` or `enums.Enum` with values to use to represent GraphQL
        enum's in Python logic.
        """
        self.name = name
        self.values = cast(dict[str, Any], getattr(values, "__members__", values))

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `EnumType` instance to the instance of GraphQL schema."""
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        graphql_type = cast(GraphQLEnumType, graphql_type)

        for key, value in self.values.items():
            if key not in graphql_type.values:
                raise ValueError(f"Value {key} is not defined on enum {self.name}")
            graphql_type.values[key].value = value

    def validate_graphql_type(self, graphql_type: GraphQLNamedType | None) -> None:
        """Validates that schema's GraphQL type associated with this `EnumType`
        is an `enum`."""
        if not graphql_type:
            raise ValueError(f"Enum {self.name} is not defined in the schema")
        if not isinstance(graphql_type, GraphQLEnumType):
            raise ValueError(
                f"{self.name} is defined in the schema, but it is instance of "
                f"{type(graphql_type).__name__} (expected {GraphQLEnumType.__name__})"
            )



import os
from typing import Any

from starlette.datastructures import UploadFile

try:
    from python_multipart.multipart import File  # type: ignore[import-untyped]
except ImportError:

    class File:
        """Mock upload file used when python-multipart is not installed."""

        file_name: bytes | None = None
        size: int = 0


def copy_args_for_tracing(value: Any) -> Any:
    if isinstance(value, dict):
        return {k: copy_args_for_tracing(v) for k, v in value.items()}
    if isinstance(value, list):
        return [copy_args_for_tracing(v) for v in value]
    if isinstance(value, UploadFile | File):
        return repr_upload_file(value)
    return value


def repr_upload_file(upload_file: UploadFile | File) -> str:
    filename: str | None | bytes
    if isinstance(upload_file, File):
        filename = upload_file.file_name
    else:
        filename = upload_file.filename

    if isinstance(filename, bytes):
        filename = filename.decode()

    mime_type: str | None

    if isinstance(upload_file, File):
        mime_type = "not/available"
    else:
        mime_type = upload_file.content_type

    if isinstance(upload_file, File):
        size = upload_file.size
    else:
        file_ = upload_file.file
        file_.seek(0, os.SEEK_END)
        size = file_.tell()
        file_.seek(0)

    return (
        f"{type(upload_file)}(mime_type={mime_type}, size={size}, filename={filename})"
    )



from collections.abc import Awaitable
from inspect import iscoroutinefunction
from typing import cast

from graphql import GraphQLNamedType
from graphql.pyutils import is_awaitable
from graphql.type import GraphQLSchema

from ariadne import InterfaceType, ObjectType
from ariadne.contrib.relay import RelayConnection
from ariadne.contrib.relay.arguments import (
    ConnectionArguments,
    ConnectionArgumentsTypeUnion,
)
from ariadne.contrib.relay.types import (
    ConnectionResolver,
    GlobalIDDecoder,
)
from ariadne.contrib.relay.utils import decode_global_id
from ariadne.types import Resolver
from ariadne.utils import is_async_callable, type_get_extension, type_set_extension


class RelayObjectType(ObjectType):
    _node_resolver: Resolver | None = None

    def __init__(
        self,
        name: str,
        connection_arguments_class: ConnectionArgumentsTypeUnion = ConnectionArguments,
    ) -> None:
        super().__init__(name)
        self.connection_arguments_class = connection_arguments_class

    def resolve_wrapper(self, resolver: ConnectionResolver):
        def wrapper(obj, info, *args, **kwargs):
            connection_arguments = self.connection_arguments_class(**kwargs)
            if is_async_callable(resolver):

                async def async_my_extension():
                    relay_connection = resolver(
                        obj, info, connection_arguments, *args, **kwargs
                    )
                    if is_awaitable(relay_connection):
                        relay_connection = await cast(
                            Awaitable[RelayConnection], relay_connection
                        )
                    return {
                        "totalCount": relay_connection.total,
                        "edges": relay_connection.get_edges(),
                        "pageInfo": relay_connection.get_page_info(
                            connection_arguments
                        ),
                    }

                return async_my_extension()

            relay_connection = resolver(
                obj, info, connection_arguments, *args, **kwargs
            )
            return {
                "totalCount": relay_connection.total,
                "edges": relay_connection.get_edges(),
                "pageInfo": relay_connection.get_page_info(connection_arguments),
            }

        return wrapper

    def connection(self, name: str):
        def decorator(resolver: ConnectionResolver) -> ConnectionResolver:
            self.set_field(name, self.resolve_wrapper(resolver))
            return resolver

        return decorator

    def node_resolver(self, resolver: Resolver):
        self._node_resolver = resolver
        return resolver

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        super().bind_to_schema(schema)

        if callable(self._node_resolver):
            graphql_type = schema.type_map.get(self.name)
            graphql_type = cast(GraphQLNamedType, graphql_type)
            type_set_extension(
                graphql_type,
                "__resolve_node__",
                self._node_resolver,
            )


class RelayNodeInterfaceType(InterfaceType):
    def __init__(
        self,
        type_resolver: Resolver | None = None,
    ) -> None:
        super().__init__("Node", type_resolver)


class RelayQueryType(RelayObjectType):
    def __init__(
        self,
        node: RelayNodeInterfaceType | None = None,
        global_id_decoder: GlobalIDDecoder = decode_global_id,
        id_field: str = "id",
    ) -> None:
        super().__init__("Query")
        if node is None:
            node = RelayNodeInterfaceType()
        self.node = node
        self.set_field("node", self.resolve_node)
        self.global_id_decoder = global_id_decoder
        self.id_field = id_field

    @property
    def bindables(self) -> tuple["RelayQueryType", "RelayNodeInterfaceType"]:
        return (self, self.node)

    def get_node_resolver(self, type_name, schema: GraphQLSchema) -> Resolver:
        type_object = schema.get_type(type_name)
        resolver: Resolver | None = None
        if type_object:
            resolver = type_get_extension(type_object, "__resolve_node__")
        if not resolver:
            raise ValueError(f"No node resolver for type {type_name}")
        return resolver

    def resolve_node(self, obj, info, *args, **kwargs):
        type_name, _ = self.global_id_decoder(kwargs[self.id_field])

        resolver = self.get_node_resolver(type_name, info.schema)

        if iscoroutinefunction(resolver):

            async def async_my_extension():
                result = await resolver(obj, info, *args, **kwargs)
                if is_awaitable(result):
                    result = await result
                return result

            return async_my_extension()
        return resolver(obj, info, *args, **kwargs)



"""Subscription handler base class for HTTP-based subscription protocols."""

from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Mapping
from logging import Logger, LoggerAdapter
from typing import Any, cast

from graphql import (
    DocumentNode,
    ExecutionResult,
    GraphQLError,
    GraphQLSchema,
    MiddlewareManager,
)
from starlette.requests import Request
from starlette.responses import Response

from ..graphql import subscribe, validate_data
from ..logger import log_error
from ..types import (
    ErrorFormatter,
    ExtensionList,
    MiddlewareList,
    QueryParser,
    QueryValidator,
    RootValue,
    ValidationRules,
)
from .events import SubscriptionEvent, SubscriptionEventType


class SubscriptionHandler(ABC):
    """Base class for HTTP-based subscription handlers.

    Provides the `generate_events` method that handles the subscription
    execution and yields `SubscriptionEvent` objects. Subclasses are used
    with `GraphQLHTTPHandler` via its `subscription_handlers` parameter.
    WebSocket-based subscriptions are handled separately by dedicated
    handlers (`GraphQLWSHandler` and `GraphQLTransportWSHandler`).
    """

    @abstractmethod
    def supports(self, request: Request, data: dict) -> bool:
        """Determine if this handler supports the given request.

        # Required arguments
        `request`: the Starlette/FastAPI request
        `data`: the parsed GraphQL request data

        # Returns
        `True` if this handler can handle the request, `False` otherwise.
        """

    @abstractmethod
    async def handle(
        self,
        request: Request,
        data: dict,
        *,
        schema: GraphQLSchema,
        context_value: Any,
        root_value: RootValue | None,
        query_parser: QueryParser | None,
        query_validator: QueryValidator | None,
        validation_rules: ValidationRules | None,
        debug: bool,
        introspection: bool,
        logger: None | str | Logger | LoggerAdapter,
        error_formatter: ErrorFormatter,
        middleware: MiddlewareList = None,
        middleware_manager_class: type[MiddlewareManager] | None = None,
        extensions: ExtensionList | None = None,
    ) -> Response:
        """Handle the subscription request."""

    async def generate_events(
        self,
        data: dict,
        *,
        schema: GraphQLSchema,
        context_value: Any,
        root_value: RootValue | None,
        query_parser: QueryParser | None,
        query_validator: QueryValidator | None,
        query_document: DocumentNode | None,
        validation_rules: ValidationRules | None,
        debug: bool,
        introspection: bool,
        logger: None | str | Logger | LoggerAdapter,
        error_formatter: ErrorFormatter,
        middleware: MiddlewareList = None,
        middleware_manager_class: type[MiddlewareManager] | None = None,
        extensions: ExtensionList | None = None,
    ) -> AsyncGenerator[SubscriptionEvent, None]:
        """Execute subscription and yield events.

        This method handles the subscription execution lifecycle:
        1. Validates the request data
        2. Calls the GraphQL subscribe function
        3. Yields SubscriptionEvent objects for each result
        4. Handles errors appropriately

        # Yields
        `SubscriptionEvent` objects for each subscription result or error.
        """
        try:
            validate_data(data)

            success, results = await subscribe(
                schema,
                data,
                context_value=context_value,
                root_value=root_value,
                query_parser=query_parser,
                query_document=query_document,
                query_validator=query_validator,
                validation_rules=validation_rules,
                debug=debug,
                introspection=introspection,
                logger=logger,
                error_formatter=error_formatter,
                middleware=middleware,
                middleware_manager_class=middleware_manager_class,
                extensions=extensions,
            )

            if not success:
                # Handle subscription setup errors
                if not isinstance(results, list):
                    error_payload = cast(list[dict[str, Any]], [results])
                else:
                    error_payload = cast(list[dict[str, Any]], results)

                yield SubscriptionEvent(
                    event_type=SubscriptionEventType.ERROR,
                    result=ExecutionResult(
                        errors=[
                            GraphQLError(
                                message=(
                                    str(error.get("message", ""))
                                    if isinstance(error, Mapping)
                                    else str(error)
                                )
                            )
                            for error in error_payload
                        ]
                    ),
                )
            else:
                results = cast(AsyncGenerator[ExecutionResult, None], results)
                try:
                    async for result in results:
                        yield SubscriptionEvent(
                            event_type=SubscriptionEventType.NEXT,
                            result=result,
                        )
                except (Exception, GraphQLError) as error:
                    if not isinstance(error, GraphQLError):
                        error = GraphQLError(str(error), original_error=error)
                        log_error(error, logger)
                    yield SubscriptionEvent(
                        event_type=SubscriptionEventType.ERROR,
                        result=ExecutionResult(errors=[error]),
                    )
        except GraphQLError as error:
            log_error(error, logger)
            yield SubscriptionEvent(
                event_type=SubscriptionEventType.ERROR,
                result=ExecutionResult(errors=[error]),
            )
        yield SubscriptionEvent(event_type=SubscriptionEventType.COMPLETE)



from .base import (
    GraphQLHandlerBase,
    GraphQLHttpHandlerBase,
    GraphQLWebsocketHandlerBase,
)
from .graphql_transport_ws import GraphQLTransportWSHandler
from .graphql_ws import GraphQLWSHandler
from .http import GraphQLHTTPHandler

__all__ = [
    "GraphQLHandlerBase",
    "GraphQLHTTPHandler",
    "GraphQLHttpHandlerBase",
    "GraphQLTransportWSHandler",
    "GraphQLWSHandler",
    "GraphQLWebsocketHandlerBase",
]



from .enums import EnumType
from .enums_default_values import (
    repair_schema_default_enum_values,
    validate_schema_default_enum_values,
)
from .executable_schema import make_executable_schema
from .extensions import ExtensionManager
from .file_uploads import combine_multipart_data, upload_scalar
from .format_error import (
    format_error,
    get_error_extension,
    get_formatted_error_context,
    get_formatted_error_traceback,
)
from .graphql import graphql, graphql_sync, subscribe
from .inputs import InputType
from .interfaces import InterfaceType
from .load_schema import load_schema_from_path
from .objects import MutationType, ObjectType, QueryType
from .resolvers import (
    is_default_resolver,
    resolve_to,
)
from .scalars import ScalarType
from .schema_names import SchemaNameConverter, convert_schema_names
from .schema_visitor import SchemaDirectiveVisitor
from .subscription_handlers import (
    SubscriptionEvent,
    SubscriptionEventType,
    SubscriptionHandler,
)
from .subscriptions import SubscriptionType
from .types import Extension, SchemaBindable
from .unions import UnionType
from .utils import (
    convert_camel_case_to_snake,
    gql,
    type_implements_interface,
    unwrap_graphql_error,
)

__all__ = [
    "EnumType",
    "Extension",
    "ExtensionManager",
    "InputType",
    "InterfaceType",
    "MutationType",
    "ObjectType",
    "QueryType",
    "ScalarType",
    "SchemaBindable",
    "SchemaDirectiveVisitor",
    "SchemaNameConverter",
    "SubscriptionEvent",
    "SubscriptionEventType",
    "SubscriptionHandler",
    "SubscriptionType",
    "UnionType",
    "combine_multipart_data",
    "convert_camel_case_to_snake",
    "convert_schema_names",
    "format_error",
    "get_error_extension",
    "get_formatted_error_context",
    "get_formatted_error_traceback",
    "gql",
    "graphql",
    "graphql_sync",
    "is_default_resolver",
    "load_schema_from_path",
    "make_executable_schema",
    "repair_schema_default_enum_values",
    "resolve_to",
    "subscribe",
    "type_implements_interface",
    "unwrap_graphql_error",
    "upload_scalar",
    "validate_schema_default_enum_values",
]



from collections import namedtuple
from collections.abc import Callable
from typing import TypeVar

from ariadne.contrib.relay.connection import RelayConnection

ConnectionResolver = TypeVar("ConnectionResolver", bound=Callable[..., RelayConnection])
GlobalIDTuple = namedtuple("GlobalIDTuple", ["type", "id"])
GlobalIDDecoder = Callable[[str], GlobalIDTuple]
GlobalIDEncoder = Callable[[str, str], str]



import inspect
from typing import Any, cast

from graphql import GraphQLError, GraphQLNamedType, GraphQLType, parse
from graphql.language import DocumentNode, OperationDefinitionNode, OperationType


def convert_camel_case_to_snake(graphql_name: str) -> str:
    """Converts a string with `camelCase` name to `snake_case`.

    Utility function used by Ariadne's name conversion logic for mapping GraphQL
    names using the `camelCase` convention to Python counterparts in `snake_case`.

    Returns a string with converted name.

    # Required arguments

    `graphql_name`: a `str` with name to convert.

    # Example

    All characters in converted name are lowercased:

    ```python
    assert convert_camel_case_to_snake("URL") == "url"
    ```

    `_` is inserted before every uppercase character that's not first and is not
    preceded by other uppercase character:

    ```python
    assert convert_camel_case_to_snake("testURL") == "test_url"
    ```

    `_` is inserted before every uppercase character succeeded by lowercased
    character:

    ```python
    assert convert_camel_case_to_snake("URLTest") == "url_test"
    ```

    `_` is inserted before every digit that's not first and is not preceded by
    other digit:

    ```python
    assert convert_camel_case_to_snake("Rfc123") == "rfc_123"
    ```
    """

    max_index = len(graphql_name) - 1
    lowered_name = graphql_name.lower()

    python_name = ""
    for i, c in enumerate(lowered_name):
        if i > 0 and (
            # testWord -> test_word
            (
                c != graphql_name[i]
                and graphql_name[i - 1] != "_"
                and python_name
                and graphql_name[i - 1] == python_name[-1]
            )
            # TESTWord -> test_word
            or (
                i < max_index
                and graphql_name[i] != c
                and graphql_name[i + 1] == lowered_name[i + 1]
                and graphql_name[i + 1] != "_"
            )
            # test134 -> test_134
            or (c.isdigit() and not graphql_name[i - 1].isdigit())
            # 134test -> 134_test
            or (
                not c.isdigit()
                and graphql_name[i] != "_"
                and graphql_name[i - 1].isdigit()
            )
        ):
            python_name += "_"
        python_name += c
    return python_name


def gql(value: str) -> str:
    """Verifies that given string is a valid GraphQL.

    Provides definition time validation for GraphQL strings. Returns unmodified
    string. Some IDEs provide GraphQL syntax for highlighting those strings.


    # Examples

    Python string in this code snippet will use GraphQL's syntax highlighting when
    viewed in VSCode:

    ```python
    type_defs = gql(
        \"\"\"
        type Query {
            hello: String!
        }
        \"\"\"
    )
    ```

    This code will raise a GraphQL parsing error:

    ```python
    type_defs = gql(
        \"\"\"
        type Query {
            hello String!
        }
        \"\"\"
    )
    ```
    """
    parse(value)
    return value


def unwrap_graphql_error(
    error: GraphQLError | Exception | None,
) -> Exception | None:
    """Recursively unwrap exception when its instance of GraphQLError.

    GraphQL query executor wraps exceptions in `GraphQLError` instances which
    contain information about exception's origin in GraphQL query or it's result.

    Original exception is available through `GraphQLError`'s `original_error`
    attribute, but sometimes `GraphQLError` can be wrapped in other `GraphQLError`.

    Returns unwrapped exception or `None` if no original exception was found.

    # Example

    Below code unwraps original `KeyError` from multiple `GraphQLError` instances:

    ```python
    error = KeyError("I am a test!")

    assert (
        unwrap_graphql_error(
            GraphQLError(
                "Error 1", GraphQLError("Error 2", GraphQLError("Error 3", original_error=error))
            )
        )
        == error
    )
    ```

    Passing other exception to `unwrap_graphql_error` results in same exception
    being returned:

    ```python
    error = ValueError("I am a test!")
    assert unwrap_graphql_error(error) == error
    ```
    """  # noqa: E501

    if isinstance(error, GraphQLError):
        return unwrap_graphql_error(error.original_error)
    return error


def type_implements_interface(interface: str, graphql_type: GraphQLType) -> bool:
    """Test if type definition from GraphQL schema implements an interface.

    Returns `True` if type implements interface and `False` if it doesn't.

    # Required arguments

    `interface`: a `str` with name of interface in GraphQL schema.

    `graphql_type`: a `GraphQLType` interface to test. It may or may not have
    the `interfaces` attribute.
    """

    try:
        return interface in [i.name for i in graphql_type.interfaces]  # type: ignore
    except AttributeError:
        pass

    return False


def get_operation_type(
    graphql_document: DocumentNode, operation_name: str | None = None
) -> OperationType:
    if operation_name:
        for d in graphql_document.definitions:
            d = cast(OperationDefinitionNode, d)
            if d.name and d.name.value == operation_name:
                return d.operation
    else:
        for definition in graphql_document.definitions:
            if isinstance(definition, OperationDefinitionNode):
                return definition.operation
    raise RuntimeError("Can't get GraphQL operation type")


def type_set_extension(
    object_type: GraphQLNamedType, extension_name: str, value: Any
) -> None:
    if getattr(object_type, "extensions", None) is None:
        object_type.extensions = {}
    object_type.extensions[extension_name] = value


def type_get_extension(
    object_type: GraphQLNamedType, extension_name: str, fallback: Any = None
) -> Any:
    return getattr(object_type, "extensions", {}).get(extension_name, fallback)


def is_async_callable(obj: Any) -> bool:
    return inspect.iscoroutinefunction(obj) or (
        callable(obj) and inspect.iscoroutinefunction(obj.__call__)
    )



"""Subscription handlers package for pluggable HTTP-based subscription protocols.

This package provides a Strategy Pattern implementation for handling GraphQL
subscriptions over HTTP-based transport protocols like SSE and HTTP callbacks.
These handlers are used with `GraphQLHTTPHandler` via its `subscription_handlers`
parameter. WebSocket-based subscriptions are handled separately by dedicated
handlers (`GraphQLWSHandler` and `GraphQLTransportWSHandler`).
"""

from .events import SubscriptionEvent, SubscriptionEventType
from .handlers import SubscriptionHandler

__all__ = [
    "SubscriptionEvent",
    "SubscriptionEventType",
    "SubscriptionHandler",
]



import os
from typing import Any

from graphql import GraphQLResolveInfo, ResponsePath
from starlette.datastructures import UploadFile

from ...resolvers import is_default_resolver

try:
    from python_multipart.multipart import File  # type: ignore[import-untyped]
except ImportError:

    class File:
        """Mock upload file used when python-multipart is not installed."""

        file_name: bytes | None = None
        size: int = 0


def copy_args_for_tracing(value: Any) -> Any:
    if isinstance(value, dict):
        return {k: copy_args_for_tracing(v) for k, v in value.items()}
    if isinstance(value, list):
        return [copy_args_for_tracing(v) for v in value]
    if isinstance(value, UploadFile | File):
        return repr_upload_file(value)
    return value


def repr_upload_file(upload_file: UploadFile | File) -> str:
    filename: str | None | bytes
    if isinstance(upload_file, File):
        filename = upload_file.file_name
    else:
        filename = upload_file.filename

    if isinstance(filename, bytes):
        filename = filename.decode()

    mime_type: str | None

    if isinstance(upload_file, File):
        mime_type = "not/available"
    else:
        mime_type = upload_file.content_type

    if isinstance(upload_file, File):
        size = upload_file.size
    else:
        file_ = upload_file.file
        file_.seek(0, os.SEEK_END)
        size = file_.tell()
        file_.seek(0)

    return (
        f"{type(upload_file)}(mime_type={mime_type}, size={size}, filename={filename})"
    )


def format_path(path: ResponsePath):
    elements = []
    current: ResponsePath | None = path
    while current:
        elements.append(current.key)
        current = current.prev
    return elements[::-1]


def should_trace(info: GraphQLResolveInfo, trace_default_resolver: bool = False):
    if info.field_name not in info.parent_type.fields:
        return False

    resolver = info.parent_type.fields[info.field_name].resolve
    if trace_default_resolver:
        default_resolver_bool = False
    else:
        default_resolver_bool = is_default_resolver(resolver)
    return not (default_resolver_bool or is_introspection_field(info))


def is_introspection_field(info: GraphQLResolveInfo):
    path = info.path
    while path:
        if isinstance(path.key, str) and is_introspection_key(path.key):
            return True
        path = path.prev
    return False


def is_introspection_key(key: str):
    return key.lower() in [
        "__schema",
        "__directive",
        "__directivelocation",
        "__type",
        "__field",
        "__inputvalue",
        "__enumvalue",
        "__typekind",
    ]  # from graphql.type.introspection.introspection_types



import asyncio
import json
import logging
from asyncio import Lock
from collections.abc import AsyncGenerator, Awaitable, Callable
from functools import partial
from http import HTTPStatus
from io import StringIO
from logging import Logger, LoggerAdapter
from typing import (
    Any,
    Literal,
    cast,
    get_args,
)

from anyio import (
    CancelScope,
    create_task_group,
    get_cancelled_exc_class,
    move_on_after,
    sleep,
)
from graphql import DocumentNode, GraphQLSchema, MiddlewareManager
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import Receive, Scope, Send

from ..asgi.handlers import GraphQLHTTPHandler
from ..exceptions import HttpError
from ..format_error import format_error
from ..graphql import (
    ExecutionResult,
    GraphQLError,
    parse_query,
    subscribe,
    validate_data,
)
from ..logger import log_error
from ..subscription_handlers.events import SubscriptionEventType
from ..subscription_handlers.handlers import SubscriptionHandler
from ..types import (
    ErrorFormatter,
    ExtensionList,
    Extensions,
    MiddlewareList,
    Middlewares,
    QueryParser,
    QueryValidator,
    RootValue,
    ValidationRules,
)

EVENT_TYPES = Literal["next", "complete"]


class GraphQLServerSentEvent:
    """GraphQLServerSentEvent is a class that represents a single Server-Sent Event
    as defined in the GraphQL SSE Protocol specification
    (https://github.com/enisdenjo/graphql-sse/blob/master/PROTOCOL.md)
    """

    DEFAULT_SEPARATOR = "\r\n"

    def __init__(
        self,
        event: EVENT_TYPES,
        result: ExecutionResult | None = None,
    ):
        """Initializes the Server-Sent Event
        # Required arguments
        `event`: the type of the event. Either "next" or "complete"

        # Optional arguments
        `result`: an `ExecutionResult` or a `dict` that represents
        the result of the operation
        """
        assert event in get_args(EVENT_TYPES), f"Invalid event type: {event}"
        self.event = event
        self.result = result
        self.logger = logging.Logger("GraphQLServerSentEvent")

    def __str__(self) -> str:
        """Returns the string representation of the Server-Sent Event"""
        buffer = StringIO()
        buffer = self._write_to_buffer(buffer, "event", self.event)
        buffer = self._write_to_buffer(
            buffer,
            "data",
            (
                self.encode_execution_result()
                if self.event == "next" and self.result
                else ""
            ),
        )
        buffer.write(self.DEFAULT_SEPARATOR)

        return buffer.getvalue()

    def _write_to_buffer(
        self, buffer: StringIO, name: str, value: str | None
    ) -> StringIO:
        """Writes an SSE field to the buffered SSE event representation

        Returns the `StringIO` buffer with the field written to it

        # Required arguments
        `buffer`: the `StringIO` buffer to write to
        `name`: the name of the field
        `value`: the value of the field
        """
        if value is not None:
            buffer.write(f"{name}: {value}{self.DEFAULT_SEPARATOR}")
        return buffer

    def encode_execution_result(self) -> str:
        """Encodes the execution result into a single line JSON string

        Returns the JSON string representation of the execution result
        """
        payload: dict[str, Any] = {}
        if self.result is not None and self.result.data is not None:
            payload["data"] = self.result.data
        if self.result is not None and self.result.errors is not None:
            errors = []
            for error in self.result.errors:
                errors.append(format_error(error))
            payload["errors"] = errors

        return json.dumps(payload)


class ServerSentEventResponse(Response):
    """Sends GraphQL SSE events using the EventSource protocol using
    Starlette's Response class based on the implementation https://github.com/sysid/sse-starlette/
    """

    # Sends a ping event to the client every 15 seconds to overcome proxy timeout issues
    DEFAULT_PING_INTERVAL = 15

    def __init__(
        self,
        *args,
        generator: AsyncGenerator[GraphQLServerSentEvent, Any],
        send_timeout: int | None = None,
        ping_interval: int | None = None,
        headers: dict[str, str] | None = None,
        **kwargs,
    ):
        """Initializes an SSE Response that sends events generated by an async generator

        # Required arguments
        `generator`: an async generator that yields `GraphQLServerSentEvent` objects

        # Optional arguments
        `send_timeout`: the timeout in seconds to send an event to the client
        `ping_interval`: the interval in seconds to send a ping event to the client,
        overrides the DEFAULT_PING_INTERVAL of 15 seconds
        `headers`: a dictionary of headers to be sent with the response
        `encoding`: the encoding to use for the response
        """
        super().__init__(*args, **kwargs)
        self.generator = generator
        self.status_code = HTTPStatus.OK
        self.send_timeout = send_timeout
        self.ping_interval = ping_interval or self.DEFAULT_PING_INTERVAL
        self.body = None

        _headers: dict[str, str] = {}
        if headers is not None:
            _headers.update(headers)
        # mandatory for servers-sent events headers
        # allow cache control header to be set by user to support fan out proxies
        # https://www.fastly.com/blog/server-sent-events-fastly
        _headers.setdefault("Cache-Control", "no-cache")
        _headers.setdefault("Connection", "keep-alive")
        _headers.setdefault("X-Accel-Buffering", "no")
        _headers.setdefault("Transfer-Encoding", "chunked")
        self.media_type = "text/event-stream"
        self.init_headers(_headers)

        self._send_lock = Lock()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """The main entrypoint for the ServerSentEventResponse
        which is called by starlette

        # Required arguments
        `scope`: the starlette Scope object

        `receive`: the starlette Receive object

        `send`: the starlette Send object

        """
        async with create_task_group() as task_group:

            async def wrap_cancelling(func: Callable[[], Awaitable[None]]) -> None:
                await func()
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap_cancelling, partial(self._ping, send))
            task_group.start_soon(wrap_cancelling, partial(self.send_events, send))
            # this will cancel the task group when the client disconnects
            await wrap_cancelling(partial(self.listen_for_disconnect, receive))

    async def _ping(self, send: Send) -> None:
        """Sends a ping event to the client every `ping_interval` seconds gets
        cancelled if the client disconnects through the anyio
        CancelScope of the TaskGroup

        # Required arguments
        `send`: the starlette Send object
        """
        while True:
            await sleep(self.ping_interval)
            async with self._send_lock:
                await send(
                    {
                        "type": "http.response.body",
                        # always encode as utf-8 as per
                        # https://html.spec.whatwg.org/multipage/server-sent-events.html#sse-processing-model
                        "body": b":\r\n\r\n",
                        "more_body": True,
                    }
                )

    async def send_events(self, send: Send) -> None:
        """Sends the events generated by the async generator to the client

        # Required arguments
        `send`: the starlette Send object

        """
        async with self._send_lock:
            await send(
                {
                    "type": "http.response.start",
                    "status": self.status_code,
                    "headers": self.raw_headers,
                }
            )

        try:
            async for event in self.generator:
                async with self._send_lock:
                    with move_on_after(self.send_timeout) as timeout:
                        await send(
                            {
                                "type": "http.response.body",
                                "body": self.encode_event(event),
                                "more_body": True,
                            }
                        )

                    if timeout.cancel_called:
                        raise asyncio.TimeoutError()

        except (get_cancelled_exc_class(),) as e:
            logging.warning(e)
        finally:
            with CancelScope(shield=True):
                async with self._send_lock:
                    await send(
                        {"type": "http.response.body", "body": b"", "more_body": False}
                    )

    @staticmethod
    async def listen_for_disconnect(receive: Receive) -> None:
        """Listens for the client disconnect event and stops the streaming by exiting
        the infinite loop. This triggers the anyio CancelScope to cancel the TaskGroup

        # Required arguments
        `receive`: the starlette Receive object
        """
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                logging.debug("Got event: http.disconnect. Stop streaming...")
                break

    @staticmethod
    def encode_event(event: GraphQLServerSentEvent) -> bytes:
        """Encodes the GraphQLServerSentEvent into a bytes object

        # Required arguments
        `event`: the GraphQLServerSentEvent object
        """
        # always encode as utf-8 as per
        # https://html.spec.whatwg.org/multipage/server-sent-events.html#sse-processing-model
        return str(event).encode("utf-8")


class SSESubscriptionHandler(SubscriptionHandler):
    """Subscription handler that delivers events via Server-Sent Events.

    Implements the `SubscriptionHandler` interface for use with
    `GraphQLHTTPHandler`'s `subscription_handlers` parameter.

    Detects SSE requests by checking for the `Accept: text/event-stream` header
    on POST requests. Uses the GraphQL SSE Protocol specification
    (https://github.com/enisdenjo/graphql-sse/blob/master/PROTOCOL.md) in
    distinct connections mode.

    # Example usage

    ```python
    from ariadne.asgi import GraphQL
    from ariadne.asgi.handlers import GraphQLHTTPHandler
    from ariadne.contrib.sse import SSESubscriptionHandler

    app = GraphQL(
        schema,
        http_handler=GraphQLHTTPHandler(
            subscription_handlers=[SSESubscriptionHandler()],
        ),
    )
    ```

    # Combining with other handlers

    ```python
    app = GraphQL(
        schema,
        http_handler=GraphQLHTTPHandler(
            subscription_handlers=[
                CallbackSubscriptionHandler(),  # Gateway access
                SSESubscriptionHandler(),  # Direct browser access
            ],
        ),
    )
    ```
    """

    def __init__(
        self,
        send_timeout: int | None = None,
        ping_interval: int | None = None,
        default_response_headers: dict[str, str] | None = None,
    ) -> None:
        """Initialize the SSE subscription handler.

        # Optional arguments

        `send_timeout`: the timeout in seconds to send an event to the client.

        `ping_interval`: the interval in seconds to send a ping event to the
        client. Defaults to `ServerSentEventResponse.DEFAULT_PING_INTERVAL`
        (15 seconds).

        `default_response_headers`: a dictionary of additional headers to be
        sent with the SSE response.
        """
        self.send_timeout = send_timeout
        self.ping_interval = ping_interval
        self.default_response_headers = default_response_headers

    def supports(self, request: Request, data: dict) -> bool:
        """Determine if this handler supports the given request.

        Returns `True` if the request is a POST with
        `Accept: text/event-stream` header.

        # Required arguments

        `request`: the Starlette/FastAPI request

        `data`: the parsed GraphQL request data
        """
        if request.method != "POST":
            return False
        accept = request.headers.get("Accept", "").split(",")
        accept = [a.strip() for a in accept]
        return "text/event-stream" in accept

    async def handle(
        self,
        request: Request,
        data: dict,
        *,
        schema: GraphQLSchema,
        context_value: Any,
        root_value: RootValue | None,
        query_parser: QueryParser | None,
        query_validator: QueryValidator | None,
        validation_rules: ValidationRules | None,
        debug: bool,
        introspection: bool,
        logger: None | str | Logger | LoggerAdapter,
        error_formatter: ErrorFormatter,
        middleware: MiddlewareList = None,
        middleware_manager_class: type[MiddlewareManager] | None = None,
        extensions: ExtensionList | None = None,
    ) -> Response:
        """Handle the subscription request via Server-Sent Events.

        Returns a `ServerSentEventResponse` that streams SSE events to the
        client with periodic ping/keepalive messages.
        """
        return ServerSentEventResponse(
            generator=self._sse_event_generator(
                data,
                schema=schema,
                context_value=context_value,
                root_value=root_value,
                query_parser=query_parser,
                query_validator=query_validator,
                validation_rules=validation_rules,
                debug=debug,
                introspection=introspection,
                logger=logger,
                error_formatter=error_formatter,
                middleware=middleware,
                middleware_manager_class=middleware_manager_class,
                extensions=extensions,
            ),
            ping_interval=self.ping_interval,
            send_timeout=self.send_timeout,
            headers=self.default_response_headers,
        )

    async def _sse_event_generator(
        self,
        data: dict,
        **kwargs: Any,
    ) -> AsyncGenerator[GraphQLServerSentEvent, None]:
        """Convert subscription events to SSE events.

        Calls `generate_events()` and maps each `SubscriptionEvent` to
        a `GraphQLServerSentEvent` for the SSE protocol.
        """
        async for event in self.generate_events(data, query_document=None, **kwargs):
            if event.event_type == SubscriptionEventType.NEXT:
                yield GraphQLServerSentEvent(event="next", result=event.result)
            elif event.event_type == SubscriptionEventType.ERROR:
                # Per GraphQL SSE Protocol: errors are sent as "next" events
                yield GraphQLServerSentEvent(event="next", result=event.result)
            elif event.event_type == SubscriptionEventType.COMPLETE:
                yield GraphQLServerSentEvent(event="complete")
            # KEEP_ALIVE is handled by ServerSentEventResponse._ping()


class GraphQLHTTPSSEHandler(GraphQLHTTPHandler):
    """Extension to the default GraphQLHTTPHandler to also handle Server-Sent Events
    as per the GraphQL SSE Protocol specification. This handler only supports the
    defined `Distinct connections mode` due to its statelessness. This implementation
    is based on the specification as of commit 80cf75b5952d1a065c95bdbd6a74304c90dbe2c5.
    For more information see the specification
    (https://github.com/enisdenjo/graphql-sse/blob/master/PROTOCOL.md)
    """

    def __init__(
        self,
        extensions: Extensions | None = None,
        middleware: Middlewares | None = None,
        middleware_manager_class: type[MiddlewareManager] | None = None,
        send_timeout: int | None = None,
        ping_interval: int | None = None,
        default_response_headers: dict[str, str] | None = None,
    ):
        super().__init__(extensions, middleware, middleware_manager_class)
        self.send_timeout = send_timeout
        self.ping_interval = ping_interval
        self.default_response_headers = default_response_headers

    async def handle_request_override(self, request: Request) -> Response | None:
        """Overrides the handle_request_override method to handle Server-Sent Events

        # Required arguments
        `request`: the starlette or FastAPI `Request` instance

        """

        if request.method == "POST":
            accept = request.headers.get("Accept", "").split(",")
            accept = [a.strip() for a in accept]
            if "text/event-stream" in accept:
                return await self.handle_sse_request(request)
        return None

    async def handle_sse_request(self, request: Request) -> Response:
        """Handles the HTTP request with GraphQL Subscription query using Server-Sent
        Events.

        # Required arguments

        `request`: the starlette `Request` instance
        """

        try:
            data = await self.extract_data_from_request(request)
            query = await self.get_query_from_sse_request(request, data)

            if self.schema is None:
                raise TypeError(
                    "schema is not set, call configure method to initialize it"
                )

            validate_data(data)
            context_value = await self.get_context_for_request(request, data)
            return ServerSentEventResponse(
                generator=self.sse_subscribe_to_graphql(query, data, context_value),
                ping_interval=self.ping_interval,
                send_timeout=self.send_timeout,
                headers=self.default_response_headers,
            )
        except (HttpError, TypeError, GraphQLError) as error:
            log_error(error, self.logger)
            if not isinstance(error, GraphQLError):
                error_message = (
                    (error.message or error.status)
                    if isinstance(error, HttpError)
                    else str(error)
                )
                error = GraphQLError(error_message, original_error=error)
            return ServerSentEventResponse(
                generator=self.sse_generate_error_response([error]),
                ping_interval=self.ping_interval,
                send_timeout=self.send_timeout,
                headers=self.default_response_headers,
            )

    async def get_query_from_sse_request(
        self, request: Request, data: Any
    ) -> DocumentNode:
        """Extracts GraphQL query from SSE request.

        Returns a `DocumentNode` with parsed query.

        # Required arguments

        `request`: the starlette `Request` instance

        `data`: an additional data parameter to potentially extract the query from
        """

        context_value = await self.get_context_for_request(request, data)
        return parse_query(context_value, self.query_parser, data)

    async def sse_subscribe_to_graphql(
        self, query_document: DocumentNode, data: Any, context_value: Any
    ):
        """Main SSE subscription generator for the GraphQL query.
        Yields `GraphQLServerSentEvent` instances and is to be consumed by a
        `ServerSentEventResponse` instance

        # Required arguments

        `query_document`: an already parsed GraphQL query.

        `data`: a `dict` with query data (`query` string, optionally `operationName`
        string and `variables` dictionary).

        `context_value`: a context value to make accessible as 'context' attribute
        of second argument (`info`) passed to resolvers and source functions.
        """

        success, results = await subscribe(
            self.schema,  # type: ignore
            data,
            context_value=context_value,
            root_value=self.root_value,
            query_document=query_document,
            query_validator=self.query_validator,
            validation_rules=self.validation_rules,
            debug=self.debug,
            introspection=self.introspection,
            logger=self.logger,
            error_formatter=self.error_formatter,
        )

        if not success:
            if not isinstance(results, list):
                error_payload = cast(list[dict[str, Any]], [results])
            else:
                error_payload = cast(list[dict[str, Any]], results)

            # This needs to be handled better, subscribe returns preformatted errors
            yield GraphQLServerSentEvent(
                event="next",
                result=ExecutionResult(
                    errors=[
                        GraphQLError(message=cast(str, error.get("message", "")))
                        for error in error_payload
                    ]
                ),
            )
        else:
            try:
                async for result in cast(AsyncGenerator, results):
                    yield GraphQLServerSentEvent(event="next", result=result)
            except (Exception, GraphQLError) as error:
                if not isinstance(error, GraphQLError):
                    error = GraphQLError(str(error), original_error=error)
                    log_error(error, self.logger)
                yield GraphQLServerSentEvent(
                    event="next", result=ExecutionResult(errors=[error])
                )

        yield GraphQLServerSentEvent(event="complete")

    @staticmethod
    async def sse_generate_error_response(
        errors: list[GraphQLError],
    ) -> AsyncGenerator[GraphQLServerSentEvent, Any]:
        """A Server-Sent Event response generator for the errors
        To be passed to a ServerSentEventResponse instance

        # Required arguments

        `errors`: a list of `GraphQLError` instances
        """

        yield GraphQLServerSentEvent(
            event="next", result=ExecutionResult(errors=errors)
        )
        yield GraphQLServerSentEvent(event="complete")



import copy
from inspect import isawaitable
from typing import Any, cast

from graphql import (
    DirectiveDefinitionNode,
    Node,
    parse,
    print_ast,
)
from graphql.language import DirectiveNode
from graphql.type import (
    GraphQLInputObjectType,
    GraphQLNamedType,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
)

from ariadne.utils import type_get_extension

_allowed_directives = [
    "skip",  # Default directive as per specs.
    "include",  # Default directive as per specs.
    "deprecated",  # Default directive as per specs.
    "external",  # Federation directive.
    "requires",  # Federation directive.
    "provides",  # Federation directive.
    "key",  # Federation directive.
    "extends",  # Federation directive.
    "link",  # Federation 2 directive.
    "shareable",  # Federation 2 directive.
    "tag",  # Federation 2 directive.
    "override",  # Federation 2 directive.
    "inaccessible",  # Federation 2 directive.
    "composeDirective",  # Federation 2.1 directive.
    "interfaceObject",  # Federation 2.3 directive.
    "authenticated",  # Federation 2.5 directive.
    "requiresScopes",  # Federation 2.5 directive.
    "policy",  # Federation 2.6 directive.
]


def _purge_directive_nodes(nodes: tuple[Node, ...]) -> tuple[Node, ...]:
    return tuple(
        node
        for node in nodes
        if not isinstance(node, DirectiveNode | DirectiveDefinitionNode)
        or node.name.value in _allowed_directives
    )


def _purge_type_directives(definition: Node) -> Node:
    """Recursively check every field defined on the Node definition
    and remove any directives found. Returns a new node with purged directives."""
    changes: dict[str, Any] = {}
    for key in definition.keys:
        value = getattr(definition, key, None)
        if isinstance(value, tuple):
            # Remove directive nodes from the tuple
            # e.g. doc -> definitions [DirectiveDefinitionNode]
            next_value = _purge_directive_nodes(cast(tuple[Node, ...], value))
            # Look for directive nodes on sub-nodes, e.g.: doc ->
            # definitions [ObjectTypeDefinitionNode] -> fields -> directives
            next_value = tuple(
                _purge_type_directives(item) if isinstance(item, Node) else item
                for item in next_value
            )
            if next_value != value:
                changes[key] = next_value
        elif isinstance(value, Node):
            new_value = _purge_type_directives(value)
            if new_value is not value:
                changes[key] = new_value
    if changes:
        new_node = copy.copy(definition)
        for key, value in changes.items():
            object.__setattr__(new_node, key, value)
        return new_node
    return definition


def purge_schema_directives(joined_type_defs: str) -> str:
    """Remove custom schema directives from federation."""
    ast_document = parse(joined_type_defs)
    ast_document = _purge_type_directives(ast_document)
    return print_ast(ast_document)


def resolve_entities(_: Any, info: GraphQLResolveInfo, **kwargs) -> Any:
    representations = list(kwargs.get("representations", []))

    result = []
    for reference in representations:
        __typename = reference["__typename"]
        type_object = info.schema.get_type(__typename)

        if not type_object or not isinstance(type_object, GraphQLObjectType):
            raise TypeError(
                f"The `_entities` resolver tried to load an entity for"
                f' type "{__typename}", but no object type of that name'
                f" was found in the schema",
            )

        resolve_reference = type_get_extension(
            type_object,
            "__resolve_reference__",
            lambda o, i, r: reference,
        )

        representation = resolve_reference(type_object, info, reference)

        if isawaitable(representation):
            result.append(add_typename_to_async_return(representation, __typename))
        else:
            result.append(add_typename_to_possible_return(representation, __typename))

    return result


async def add_typename_to_async_return(obj: Any, typename: str) -> Any:
    return add_typename_to_possible_return(await obj, typename)


def get_entity_types(schema: GraphQLSchema) -> list[GraphQLNamedType]:
    """Get all types that include the @key directive."""
    schema_types = schema.type_map.values()

    def check_type(t):
        return isinstance(t, GraphQLObjectType) and includes_directive(t, "key")

    return [t for t in schema_types if check_type(t)]


def includes_directive(
    type_object: GraphQLNamedType,
    directive_name: str,
) -> bool:
    """Check if specified type includes a directive."""
    if isinstance(type_object, GraphQLInputObjectType):
        return False

    directives = gather_directives(type_object)
    return any(d.name.value == directive_name for d in directives)


def gather_directives(
    type_object: GraphQLNamedType,
) -> list[DirectiveNode]:
    """Get all directive attached to a type."""
    directives: list[DirectiveNode] = []

    if hasattr(type_object, "extension_ast_nodes") and type_object.extension_ast_nodes:
        for ast_node in type_object.extension_ast_nodes:
            if ast_node.directives:
                directives.extend(ast_node.directives)

    if (
        hasattr(type_object, "ast_node")
        and type_object.ast_node
        and type_object.ast_node.directives
    ):
        directives.extend(type_object.ast_node.directives)

    return directives


def add_typename_to_possible_return(obj: Any, typename: str) -> Any:
    if obj is not None:
        if isinstance(obj, dict):
            obj["__typename"] = typename
        else:
            setattr(obj, f"_{obj.__class__.__name__}__typename", typename)
        return obj
    return None



"""Subscription event abstraction for HTTP-based subscription handlers."""

from dataclasses import dataclass
from enum import Enum

from graphql import ExecutionResult


class SubscriptionEventType(Enum):
    """Type of subscription event."""

    NEXT = "next"
    ERROR = "error"
    COMPLETE = "complete"
    KEEP_ALIVE = "keep_alive"


@dataclass
class SubscriptionEvent:
    """Represents a subscription event for HTTP-based subscription handlers.

    This is a representation of subscription events used by `SubscriptionHandler`
    subclasses to deliver subscription data over HTTP-based protocols
    (SSE, HTTP callbacks, etc.). These events are used with `GraphQLHTTPHandler`
    and are not related to WebSocket-based subscription handlers.

    # Attributes

    `event_type`: the type of the event (next, error, complete, keep_alive)

    `result`: an optional `ExecutionResult` containing the subscription data or errors
    """

    event_type: SubscriptionEventType
    result: ExecutionResult | None = None



from typing_extensions import (
    TypeAliasType,  # TODO: Replace with typing.TypeAliasType while dropping 3.11 support.
)


class ForwardConnectionArguments:
    first: int | None
    after: str | None

    def __init__(self, *, first: int | None = None, after: str | None = None) -> None:
        self.first = first
        self.after = after


class BackwardConnectionArguments:
    last: int | None
    before: str | None

    def __init__(self, *, last: int | None = None, before: str | None = None) -> None:
        self.last = last
        self.before = before


class ConnectionArguments:
    def __init__(
        self,
        *,
        first: int | None = None,
        after: str | None = None,
        last: int | None = None,
        before: str | None = None,
    ) -> None:
        self.first = first
        self.after = after
        self.last = last
        self.before = before


ConnectionArgumentsUnion = TypeAliasType(
    "ConnectionArgumentsUnion",
    ForwardConnectionArguments | BackwardConnectionArguments | ConnectionArguments,
)
ConnectionArgumentsTypeUnion = TypeAliasType(
    "ConnectionArgumentsTypeUnion",
    type[ForwardConnectionArguments]
    | type[BackwardConnectionArguments]
    | type[ConnectionArguments],
)



import inspect
from collections.abc import Callable
from typing import Any, cast

try:
    from anyio import to_thread
except ImportError:
    import asyncio

    class _ThreadAdapter:
        @staticmethod
        async def run_sync(func: Callable[..., Any], *args: Any) -> Any:
            return await asyncio.to_thread(func, *args)

    to_thread: _ThreadAdapter = _ThreadAdapter()

from graphql.type import GraphQLSchema

from .objects import ObjectType
from .types import Subscriber


class SubscriptionType(ObjectType):
    """Bindable populating the Subscription type in a GraphQL schema with Python logic.

    Extends `ObjectType`, providing `source` decorator and `set_source` method, used
    to set subscription sources for it's fields.

    # Subscription sources ("subscribers")

    Subscription source is a function that is a generator (async or sync).
    This function is used to subscribe to source of events or messages.
    It can also filter the messages by not yielding them.

    Its signature is same as resolver:

    ```python
    # Async generator (executed directly)
    async def source_fn(root_value: Any, info: GraphQLResolveInfo, **field_args) -> Any:
        yield ...


    # Synchronous generator (executed in a worker thread to avoid blocking)
    def source_fn(root_value: Any, info: GraphQLResolveInfo, **field_args) -> Any:
        yield ...
    ```

    # Subscription resolvers

    Subscription resolvers are called with message returned from the source. Their role
    is to convert this message into Python representation of a type associated with
    subscription's field in GraphQL schema. Its called with message yielded from
    source function as first argument.

    ```python
    def resolver_fn(message: Any, info: GraphQLResolveInfo, **field_args) -> Any:
        return ...
    ```

    # GraphQL arguments

    When subscription field has arguments those arguments values are passed
    to both source and resolver functions.

    # Example source and resolver

    ```python
    from ariadne import SubscriptionType, make_executable_schema
    from broadcast import broadcast

    from .models import Post


    subscription_type = SubscriptionType()


    @subscription_type.source("post")
    async def source_post(*_, category: Optional[str] = None) -> dict:
        async with broadcast.subscribe(channel="NEW_POSTS") as subscriber:
            async for event in subscriber:
                message = json.loads(event.message)
                if not category or message["category"] == category:
                    yield message


    @subscription_type.field("post")
    async def resolve_post(message: dict, *_, category: Optional[str] = None) -> Post:
        return await Post.get_one(id=message["post_id"])
    ```

    # Synchronous generator example

    Synchronous generators are automatically executed in worker threads to avoid
    blocking the event loop. **Note**: This thread-based implementation has
    scalability implications:
    - Limited number of concurrent subscriptions (bounded by thread pool size)
    - Higher memory overhead per subscription (thread stack + generator state)
    - Performance overhead from thread context switching

    For high-concurrency scenarios, prefer async generators when possible.

    ```python
    from ariadne import SubscriptionType, make_executable_schema
    import time

    subscription_type = SubscriptionType()


    @subscription_type.source("timeUpdates")
    def source_time(*_, interval: int = 1) -> dict:
        while True:
            yield {"timestamp": time.time()}
            time.sleep(interval)


    @subscription_type.field("timeUpdates")
    def resolve_time(message: dict, *_) -> dict:
        return message


    schema = make_executable_schema(
        \"\"\"
        type Query {
            _: Boolean
        }
        type Subscription {
            timeUpdates(interval: Int): TimeUpdate!
        }
        type TimeUpdate {
            timestamp: Float!
        }
        \"\"\",
        subscription_type
    )


    schema = make_executable_schema(
        \"\"\"
        type Query {
            \"Valid schema must define the Query type\"
            none: Int
        }

        type Subscription {
            post(category: ID): Post!
        }

        type Post {
            id: ID!
            author: String!
            text: String!
        }
        \"\"\",
        subscription_type
    )
    ```

    # Example chat

    [Ariadne GraphQL Chat Example](https://github.com/mirumee/ariadne-graphql-chat-example)
    is the Github repository with GraphQL chat application, using Redis message backend,
    Broadcaster library for publishing and subscribing to messages and React.js client
    using Apollo-Client subscriptions.
    """

    _subscribers: dict[str, Subscriber]

    def __init__(self) -> None:
        """Initializes the `SubscriptionType` with a GraphQL name set to `Subscription`."""  # noqa: E501
        super().__init__("Subscription")
        self._subscribers = {}

    def source(self, name: str) -> Callable[[Subscriber], Subscriber]:
        """Return a decorator that sets decorated function as a source for named field.

        Wrapper for `create_register_subscriber` that on runtime
        validates `name` to be a string.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        bind decorated source to.
        """
        if not isinstance(name, str):
            raise ValueError(
                'source decorator should be passed a field name: @foo.source("name")'
            )
        return self.create_register_subscriber(name)

    def create_register_subscriber(
        self, name: str
    ) -> Callable[[Subscriber], Subscriber]:
        """Return a decorator that sets decorated function as a source for named field.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        bind decorated source to.

        # Synchronous generators

        If the decorated function is a synchronous generator (`def` + `yield`), it will
        be automatically wrapped in an async generator that executes the sync generator
        in a worker thread to avoid blocking the event loop. The sync generator will
        be properly cleaned up when the subscription ends or client disconnects.

        **Scalability considerations**: Thread-based execution limits concurrent
        subscriptions (bounded by thread pool size), increases memory overhead per
        subscription, and adds performance overhead from thread context switching.
        For high-concurrency scenarios, prefer async generators.
        """

        def register_subscriber(generator: Subscriber) -> Subscriber:
            is_sync_generator = inspect.isgeneratorfunction(
                generator
            ) and not inspect.isasyncgenfunction(generator)

            if is_sync_generator:

                async def async_wrapper(*args, **kwargs):
                    sync_gen = generator(*args, **kwargs)

                    try:
                        while True:
                            try:

                                def get_next(gen: Any) -> Any:
                                    return next(gen)

                                value = await to_thread.run_sync(get_next, sync_gen)
                                yield value
                            except StopIteration:
                                break
                            except RuntimeError as e:
                                if isinstance(
                                    e.__cause__, StopIteration
                                ) or "StopIteration" in str(e):
                                    break
                                raise
                    finally:
                        try:
                            if hasattr(sync_gen, "close"):
                                close_fn = cast(Callable[[], None], sync_gen.close)
                                await to_thread.run_sync(close_fn)
                        except (Exception, GeneratorExit):
                            pass

                wrapped: Subscriber = async_wrapper
                self._subscribers[name] = wrapped
                return wrapped
            else:
                self._subscribers[name] = generator
                return generator

        return register_subscriber

    def set_source(self, name, generator: Subscriber) -> Subscriber:
        """Set a source for the field name.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        set this source for.

        `generator`: a `Subscriber` function to use as an source. Can be either an async
        generator or a synchronous generator. Synchronous generators will be
        automatically wrapped to run in a worker thread.

        **Scalability note**: Sync generators use thread-based execution which limits
        concurrent subscriptions and increases memory/performance overhead compared
        to async generators.
        """
        is_sync_generator = inspect.isgeneratorfunction(
            generator
        ) and not inspect.isasyncgenfunction(generator)

        if is_sync_generator:

            async def async_wrapper(*args, **kwargs):
                sync_gen = generator(*args, **kwargs)

                try:
                    while True:
                        try:

                            def get_next(gen: Any) -> Any:
                                return next(gen)

                            value = await to_thread.run_sync(get_next, sync_gen)
                            yield value
                        except StopIteration:
                            break
                        except RuntimeError as e:
                            if isinstance(
                                e.__cause__, StopIteration
                            ) or "StopIteration" in str(e):
                                break
                            raise
                finally:
                    try:
                        if hasattr(sync_gen, "close"):
                            close_fn = cast(Callable[[], None], sync_gen.close)
                            await to_thread.run_sync(close_fn)
                    except (Exception, GeneratorExit):
                        pass

            wrapped: Subscriber = async_wrapper
            self._subscribers[name] = wrapped
            return wrapped
        else:
            self._subscribers[name] = generator
            return generator

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `SubscriptionType` instance to the instance of GraphQL schema.

        If it has any previously set subscription resolvers or source functions,
        those will be replaced with new ones from this instance.
        """
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        self.bind_resolvers_to_graphql_type(graphql_type)
        self.bind_subscribers_to_graphql_type(graphql_type)

    def bind_subscribers_to_graphql_type(self, graphql_type):
        """Binds this `SubscriptionType` instance's source functions.

        Source functions are set to fields `subscribe` attributes.
        """
        for field, subscriber in self._subscribers.items():
            if field not in graphql_type.fields:
                raise ValueError(f"Field {field} is not defined on type {self.name}")

            graphql_type.fields[field].subscribe = subscriber



from base64 import b64decode, b64encode

from ariadne.contrib.relay.types import (
    GlobalIDTuple,
)


def decode_global_id(gid: str) -> GlobalIDTuple:
    return GlobalIDTuple(*b64decode(gid).decode().split(":"))


def encode_global_id(type_name: str, _id: str) -> str:
    return b64encode(f"{type_name}:{_id}".encode()).decode()



from collections.abc import Mapping
from typing import Any

from graphql import default_field_resolver
from graphql.type import (
    GraphQLResolveInfo,
)

from .types import Resolver


def resolve_parent_field(parent: Any, field_name: str) -> Any:
    value = parent
    for name in field_name.split("."):
        if isinstance(value, Mapping):
            value = value.get(name)
        else:
            value = getattr(value, name, None)

        if value is None:
            break
    return value


def resolve_to(attr_name: str) -> Resolver:
    """Create a resolver that resolves to given attribute or dict key.

    Returns a resolver function that can be used as resolver.

    Usually not used directly  but through higher level features like aliases
    or schema names conversion.

    # Required arguments

    `attr_name`: a `str` with name of attribute or `dict` key to return from
    resolved object.
    """

    def resolver(parent: Any, info: GraphQLResolveInfo, **kwargs) -> Any:
        value = resolve_parent_field(parent, attr_name)
        if callable(value):
            return value(info, **kwargs)
        return value

    resolver._ariadne_alias_resolver = True  # type: ignore
    return resolver


def is_default_resolver(resolver: Resolver | None) -> bool:
    """Test if resolver function is default resolver implemented by
    `graphql-core` or Ariadne.

    Returns `True` if resolver function is `None`, `graphql.default_field_resolver`
    or was created by Ariadne's `resolve_to` utility. Returns `False` otherwise.

    `True` is returned for `None` because query executor defaults to the
    `graphql.default_field_resolver` is there's no resolver function set on a
    field.

    # Required arguments

    `resolver`: a function `None` to test or `None`.
    """
    if not resolver or resolver == default_field_resolver:
        return True
    return hasattr(resolver, "_ariadne_alias_resolver")



from __future__ import annotations

import json
from http import HTTPStatus
from inspect import isawaitable
from typing import TYPE_CHECKING, Any, cast

from graphql import DocumentNode, MiddlewareManager
from starlette.datastructures import UploadFile
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, PlainTextResponse, Response
from starlette.types import Receive, Scope, Send

from ...constants import (
    DATA_TYPE_JSON,
    DATA_TYPE_MULTIPART,
)
from ...exceptions import HttpBadRequestError, HttpError
from ...explorer import Explorer
from ...file_uploads import combine_multipart_data
from ...graphql import graphql
from ...types import (
    ContextValue,
    ExtensionList,
    Extensions,
    GraphQLResult,
    MiddlewareList,
    Middlewares,
)
from .base import GraphQLHttpHandlerBase

if TYPE_CHECKING:
    from ...subscription_handlers import SubscriptionHandler


class GraphQLHTTPHandler(GraphQLHttpHandlerBase):
    """Default ASGI handler for HTTP requests.

    Supports the `Query` and `Mutation` operations. Can also handle
    subscriptions when configured with `subscription_handlers`.
    """

    def __init__(
        self,
        extensions: Extensions | None = None,
        middleware: Middlewares | None = None,
        middleware_manager_class: type[MiddlewareManager] | None = None,
        subscription_handlers: list[SubscriptionHandler] | None = None,
    ) -> None:
        """Initializes the HTTP handler.

        # Optional arguments

        `extensions`: an `Extensions` list or callable returning a
        list of extensions server should use during query execution. Defaults
        to no extensions.

        `middleware`: a `Middlewares` list or callable returning a list of
        middlewares server should use during query execution. Defaults to no
        middlewares.

        `middleware_manager_class`: a `MiddlewareManager` type or subclass to
        use for combining provided middlewares into single wrapper for resolvers
        by the server. Defaults to `graphql.MiddlewareManager`. Is only used
        if `extensions` or `middleware` options are set.

        `subscription_handlers`: a list of `SubscriptionHandler` instances to
        handle GraphQL subscriptions. Handlers are tried in order; the first
        handler whose `supports()` method returns `True` handles the request.
        """
        super().__init__()

        self.extensions = extensions
        self.middleware = middleware
        self.middleware_manager_class = middleware_manager_class or MiddlewareManager
        self.subscription_handlers: list[SubscriptionHandler] = (
            subscription_handlers or []
        )

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        """An entrypoint for the GraphQL HTTP handler.

        This method is called by the Ariadne ASGI GraphQL application to execute
        queries done using the HTTP protocol.

        It creates the `starlette.requests.Request` instance, calls
        `handle_request` method with it, then sends the response back to the client.

        # Required arguments

        `scope`: The connection scope information, a dictionary that contains
        at least a type key specifying the protocol that is incoming.

        `receive`: an awaitable callable that will yield a new event dictionary
        when one is available.

        `send`: an awaitable callable taking a single event dictionary as a
        positional argument that will return once the send has been completed
        or the connection has been closed.

        Details about the arguments and their usage are described in the
        ASGI specification:

        https://asgi.readthedocs.io/en/latest/specs/main.html
        """
        request = Request(scope=scope, receive=receive)
        response = await self.handle_request(request)
        await response(scope, receive, send)

    async def handle_request_override(self, request: Request) -> Response | None:
        """Override the default request handling logic in subclasses.
        Is called in the `handle_request` method before the default logic.
        If None is returned, the default logic is executed.

         # Required arguments:
         `request`: the `Request` instance from Starlette or FastAPI.
        """
        return None

    async def handle_request(self, request: Request) -> Response:
        """Handle GraphQL request and return a response for the client.

        Is called by the `handle` method and `handle_request` method of the
        ASGI GraphQL application.

        Handles three HTTP methods:

        `GET`: returns GraphQL explorer or 405 error response if explorer or
        introspection is disabled.

        `POST`: executes the GraphQL query from either `application/json` or
        `multipart/form-data` requests.

        `OPTIONS`: returns supported HTTP methods.

        Returns Starlette's `Response` instance, which is also works in FastAPI.

        # Required arguments:

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        response = await self.handle_request_override(request)
        if response is not None:
            return response

        if request.method == "GET":
            if self.execute_get_queries and request.query_params.get("query"):
                return await self.graphql_http_server(request)
            if self.introspection and self.explorer:
                # only render explorer when introspection is enabled
                return await self.render_explorer(request, self.explorer)

        if request.method == "POST":
            return await self.graphql_http_server(request)

        return self.handle_not_allowed_method(request)

    async def render_explorer(self, request: Request, explorer: Explorer) -> Response:
        """Return an HTML response with GraphQL explorer.

        # Required arguments:

        `request`: the `Request` instance from Starlette or FastAPI.

        `explorer`: an `Explorer` instance that implements the
        `html(request: Request)` method which returns either the `str` with HTML
        or `None`. If the explorer returns `None`, `405` method not allowed response
        is returned instead.
        """
        explorer_html = explorer.html(request)
        if isawaitable(explorer_html):
            explorer_html = await explorer_html
        if explorer_html:
            return HTMLResponse(explorer_html)

        return self.handle_not_allowed_method(request)

    async def graphql_http_server(self, request: Request) -> Response:
        """Handles the HTTP request with GraphQL query.

        Extracts GraphQL query data from requests and then executes it using
        the `execute_graphql_query` method. If subscription handlers are
        configured and one of them supports the request, delegates to that
        handler instead.

        Returns the JSON response from Starlette, or a response from a
        subscription handler.

        If the request's data was invalid or missing, a plaintext response with an
        error message and 400 status code is returned instead.

        # Required arguments:

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        try:
            data = await self.extract_data_from_request(request)
        except HttpError as error:
            return PlainTextResponse(
                error.message or error.status, status_code=HTTPStatus.BAD_REQUEST
            )

        # Check if any subscription handler can handle this request
        if self.subscription_handlers and isinstance(data, dict) and self.schema:
            for handler in self.subscription_handlers:
                if handler.supports(request, data):
                    context_value = await self.get_context_for_request(request, data)
                    extensions = await self.get_extensions_for_request(
                        request, context_value
                    )
                    middleware = await self.get_middleware_for_request(
                        request, context_value
                    )
                    return await handler.handle(
                        request,
                        data,
                        schema=self.schema,
                        context_value=context_value,
                        root_value=self.root_value,
                        query_parser=self.query_parser,
                        query_validator=self.query_validator,
                        validation_rules=self.validation_rules,
                        debug=self.debug,
                        introspection=self.introspection,
                        logger=self.logger,
                        error_formatter=self.error_formatter,
                        extensions=extensions,
                        middleware=middleware,
                        middleware_manager_class=self.middleware_manager_class,
                    )

        success, result = await self.execute_graphql_query(request, data)
        return await self.create_json_response(request, result, success)

    async def extract_data_from_request(self, request: Request) -> Any:
        """Extracts GraphQL request data from request.

        Returns GraphQL query data that was not yet validated.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        content_type = request.headers.get("Content-Type", "")
        content_type = content_type.split(";")[0]

        if content_type == DATA_TYPE_JSON:
            return await self.extract_data_from_json_request(request)
        if content_type == DATA_TYPE_MULTIPART:
            return await self.extract_data_from_multipart_request(request)
        if (
            request.method == "GET"
            and self.execute_get_queries
            and request.query_params.get("query")
        ):
            return self.extract_data_from_get_request(request)

        raise HttpBadRequestError(
            f"Posted content must be of type {DATA_TYPE_JSON} or {DATA_TYPE_MULTIPART}"
        )

    async def extract_data_from_json_request(self, request: Request) -> Any:
        """Extracts GraphQL data from JSON request.

        Returns GraphQL query data that was not yet validated.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        try:
            return await request.json()
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError("Request body is not a valid JSON") from ex

    async def extract_data_from_multipart_request(
        self, request: Request
    ) -> dict | list:
        """Extracts GraphQL data from `multipart/form-data` request.

        Returns an unvalidated `dict` or `list` with GraphQL query data.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        try:
            request_body = await request.form()
        except ValueError as ex:
            raise HttpBadRequestError(
                "Request body is not a valid multipart/form-data"
            ) from ex

        try:
            operations = json.loads(cast(Any, request_body.get("operations")))
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError(
                "Request 'operations' multipart field is not a valid JSON"
            ) from ex
        try:
            files_map = json.loads(cast(Any, request_body.get("map")))
        except (TypeError, ValueError) as ex:
            raise HttpBadRequestError(
                "Request 'map' multipart field is not a valid JSON"
            ) from ex

        request_files = {
            key: value
            for key, value in request_body.items()
            if isinstance(value, UploadFile)
        }

        return combine_multipart_data(operations, files_map, request_files)

    def extract_data_from_get_request(self, request: Request) -> dict:
        """Extracts GraphQL data from GET request's querystring.

        Returns a `dict` with GraphQL query data that was not yet validated.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        query = request.query_params["query"].strip()
        operation_name = request.query_params.get("operationName", "").strip()
        variables = request.query_params.get("variables", "").strip()

        clean_variables = None

        if variables:
            try:
                clean_variables = json.loads(variables)
            except (TypeError, ValueError) as ex:
                raise HttpBadRequestError(
                    "Variables query arg is not a valid JSON"
                ) from ex

        return {
            "query": query,
            "operationName": operation_name or None,
            "variables": clean_variables,
        }

    async def execute_graphql_query(
        self,
        request: Any,
        data: Any,
        *,
        context_value: Any = None,
        query_document: DocumentNode | None = None,
    ) -> GraphQLResult:
        """Executes GraphQL query from `request` and returns `GraphQLResult`.

        Creates GraphQL `ContextValue`, initializes extensions and middlewares,
        then runs the `graphql` function from Ariadne to execute the query.

        # Requires arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `data`: a GraphQL data.

        # Optional arguments

        `context_value`: a `ContextValue` for this request.

        `query_document`: an already parsed GraphQL query. Setting this option
        will prevent `graphql` from parsing `query` string from `data` second time.
        """
        if context_value is None:
            context_value = await self.get_context_for_request(request, data)

        extensions = await self.get_extensions_for_request(request, context_value)
        middleware = await self.get_middleware_for_request(request, context_value)

        if self.schema is None:
            raise TypeError("schema is not set, call configure method to initialize it")

        if isinstance(request, Request):
            require_query = request.method == "GET"
        else:
            require_query = False

        return await graphql(
            self.schema,
            data,
            context_value=context_value,
            root_value=self.root_value,
            query_parser=self.query_parser,
            query_validator=self.query_validator,
            query_document=query_document,
            validation_rules=self.validation_rules,
            require_query=require_query,
            debug=self.debug,
            introspection=self.introspection,
            logger=self.logger,
            error_formatter=self.error_formatter,
            extensions=extensions,
            middleware=middleware,
            middleware_manager_class=self.middleware_manager_class,
            execution_context_class=self.execution_context_class,
        )

    async def get_extensions_for_request(
        self, request: Any, context: ContextValue | None
    ) -> ExtensionList:
        """Returns extensions to use when handling the GraphQL request.

        Returns `ExtensionList`, a list of extensions to use or `None`.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `context`: a `ContextValue` for this request.
        """
        if callable(self.extensions):
            extensions = self.extensions(request, context)  # ty: ignore
            if isawaitable(extensions):
                extensions = await extensions
            return cast(ExtensionList, extensions)
        return self.extensions

    async def get_middleware_for_request(
        self, request: Any, context: ContextValue | None
    ) -> MiddlewareList:
        """Returns GraphQL middlewares to use when handling the GraphQL request.

        Returns `MiddlewareList`, a list of middlewares to use or `None`.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `context`: a `ContextValue` for this request.
        """
        middleware = self.middleware
        if callable(middleware):
            middleware = middleware(request, context)  # ty: ignore
            if isawaitable(middleware):
                middleware = await middleware
        if middleware:
            return cast(MiddlewareList, middleware)
        return None

    async def create_json_response(
        self,
        request: Request,
        result: dict,
        success: bool,
    ) -> Response:
        """Creates JSON response from GraphQL's query result.

        Returns Starlette's `JSONResponse` instance that's also compatible
        with FastAPI. If `success` is `True`, response's status code is 200.
        Status code 400 is used otherwise.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.

        `result`: a JSON-serializable `dict` with query result.

        `success`: a `bool` specifying if
        """
        status_code = HTTPStatus.OK if success else HTTPStatus.BAD_REQUEST
        return JSONResponse(result, status_code=status_code)

    def handle_not_allowed_method(self, request: Request):
        """Handles request for unsupported HTTP method.

        Returns 200 response for `OPTIONS` request and 405 response for other
        methods. All responses have empty body.

        # Required arguments

        `request`: the `Request` instance from Starlette or FastAPI.
        """
        allowed_methods = ["OPTIONS", "POST"]
        if self.introspection:
            allowed_methods.append("GET")
        allow_header = {"Allow": ", ".join(allowed_methods)}

        if request.method == "OPTIONS":
            return Response(headers=allow_header)

        return Response(status_code=HTTPStatus.METHOD_NOT_ALLOWED, headers=allow_header)



from enum import Enum
from http import HTTPStatus

DATA_TYPE_JSON = "application/json"
DATA_TYPE_MULTIPART = "multipart/form-data"

CONTENT_TYPE_JSON = "application/json; charset=UTF-8"
CONTENT_TYPE_TEXT_HTML = "text/html; charset=UTF-8"
CONTENT_TYPE_TEXT_PLAIN = "text/plain; charset=UTF-8"


class HttpStatusResponse(Enum):
    OK = f"{HTTPStatus.OK} OK"
    BAD_REQUEST = f"{HTTPStatus.BAD_REQUEST} BAD REQUEST"
    METHOD_NOT_ALLOWED = f"{HTTPStatus.METHOD_NOT_ALLOWED} METHOD NOT ALLOWED"
